---@class Component : Object
---@field protected __class string
---@field private _root LuaGuiElement
Component = Object:extendAs("model.Component")

---[STATIC]
---@protected
---@param parent LuaGuiElement
---@return self
function Component:create(parent)
    error("Not implemented")
end

---[STATIC]
---@protected
---@param root LuaGuiElement
---@return self
function Component:new(root)
    local this = Component:super(Object:new())
    this._root = assert(root)
    return this
end

---@public
---@param event Event
function Component:handle(event)
    for _, child in ipairs(self:allChildren()) do
        child:handle(event)
    end
end

---@public
---@param guiEvent GuiEvent
function Component:handleGuiEvent(guiEvent)
    local child = self:findChildClassContaining(guiEvent:element())
    if child then
        child:handleGuiEvent(guiEvent)
    end
end

---@public
function Component:lock()
    for _, child in ipairs(self:allChildren()) do
        child:lock()
    end
end

---@public
function Component:unlock()
    for _, child in ipairs(self:allChildren()) do
        child:unlock()
    end
end

---@generic T : Component
---@param class T
---@return T
function Component:child(class)
    return self:children(class)[1]
end

---@generic T : Component
---@param class T
---@return T[]
function Component:children(class)
    local children = {}
    for _, child in ipairs(self:allChildren()) do
        if child:isInstanceOf(class) then
            table.insert(children, child)
        end
    end
    return children
end

---@public
---@return Component[]
function Component:allChildren()
    local children = {}
    for _, class in ipairs(self:childrenClasses()) do
        for _, child in ipairs(class:findChildren(self:root())) do
            table.insert(children, child)
        end
    end
    return children
end

---@return Component[]
function Component:childrenClasses()
    return {}
end

---@return self[]
function Component:siblingsAndMe()
    return self:findChildren(self:parent())
end

---@generic T : Component
---@param class T
---@return T
function Component:ancestor(class)
    return class:findUpFrom(self:root())
end

---@protected
---@param replacement Component
function Component:replaceWith(replacement)
    self:parent().swap_children(self:index(), replacement:index())
    self:delete()
end

---@protected
---@param other Component
function Component:swapWith(other)
    self:parent().swap_children(self:index(), other:index())
end

---@public
---@return boolean
function Component:isValid()
    return self:root().valid
end

---@public
function Component:show()
    self:root().visible = true
end

---@public
function Component:hide()
    self:root().visible = false
end

---@public
function Component:delete()
    self:root().destroy()
end

---@protected
---@return number
function Component:index()
    return self:root().get_index_in_parent()
end

---@protected
---@return LuaGuiElement
function Component:parent()
    return self:root().parent
end

---@protected
---@return LuaGuiElement
function Component:root()
    return self._root
end

-- [STATIC]

---[STATIC]
---Sets element.tags.class to value of self.__class to recognize it while searching
---@protected
---@param parent LuaGuiElement
---@param addParameters LuaGuiElement.add_parameters
---@return LuaGuiElement
function Component:createRoot(parent, addParameters)
    addParameters.tags = { className = self:className() }
    return parent.add(addParameters)
end

---[STATIC]
---@public
---@param element LuaGuiElement
---@return boolean
function Component:is(element)
    return element.tags and element.tags.className == self:className()
end

---@private
---@param element LuaGuiElement
---@return self[]
function Component:findChildren(element)
    if element.children then
        local founded = {}
        for _, child in ipairs(element.children) do
            if self:is(child) then
                table.insert(founded, self:new(child))
            end
        end
        return founded
    else
        return {}
    end
end

---[STATIC]
---@public
---@return boolean
---@param guiEvent GuiEvent
function Component:isOriginOf(guiEvent)
    return self:is(guiEvent:element())
end

---[STATIC]
---@public
---@return boolean
---@param guiEvent GuiEvent
function Component:containsOriginOf(guiEvent)
    return self:contains(guiEvent:element())
end

---[STATIC]
---@private
---@param element LuaGuiElement
function Component:findChildClassContaining(element)
    for _, child in ipairs(self:childrenClasses()) do
        if child:contains(element) then
            return child:findUpFrom(element)
        end
    end
end

---[STATIC]
---@private
---@param element LuaGuiElement
---@return boolean
function Component:contains(element)
    return self:findUpFrom(element) ~= nil
end

---[STATIC]
---Recognizes itself by element.tags.class set in Component:createRootElement
---@public
---@param element LuaGuiElement
---@return self
function Component:findUpFrom(element)
    -- is root LuaGuiElement (Screen, etc.)
    if element.parent == nil then
        return nil
    elseif self:is(element) then
        return self:new(element)
    else
        return self:findUpFrom(element.parent)
    end
end
