import('model.Component')
import('model.toolbar.header.ToolbarHeader')
import('model.toolbar.ToolbarContent')
import('model.toolbar.header.Unlock')
import('model.toolbar.SectionList')
import('model.toolbar.section.Section')
import('model.toolbar.section.table.Table')

---@class Toolbar : Component
Toolbar = Component:extendAs("model.toolbar.Toolbar")

function Toolbar:create(parent)
    local root = Toolbar:createRoot(parent, {
        type = "frame",
        direction = "vertical",
        style = "toolbar" })
    ToolbarHeader:create(root)
    ToolbarContent:create(root)
    local toolbar = Toolbar:new(root)
    toolbar:adjustSize()
    return toolbar
end

function Toolbar:new(root)
    return Toolbar:super(Component:new(root))
end

function Toolbar:childrenClasses()
    return { ToolbarHeader, ToolbarContent }
end

---@return ToolbarHeader
function Toolbar:header()
    return self:child(ToolbarHeader)
end

---@return ToolbarContent
function Toolbar:content()
    return self:child(ToolbarContent)
end

---@public
function Toolbar:centerOnScreen()
    self:root().force_auto_center()
end

function Toolbar:handleGuiEvent(guiEvent)
    if Toolbar:isOriginOf(guiEvent) then
        if guiEvent:isLocationChange() then
            self:keepWithinTheScreen()
            self:turnOffCenteringOnScreen()
        end
    else
        Component.handleGuiEvent(self, guiEvent)
    end
end

---@private
function Toolbar:keepWithinTheScreen()
    local location = self:location()
    if location.x < 0 then
        location.x = 0
    end
    if location.y < 0 then
        location.y = 0
    end
    self:setLocation(location)
end

---@private
function Toolbar:turnOffCenteringOnScreen()
    self:root().auto_center = false
end

---@private
---@param location GuiLocation
function Toolbar:setLocation(location)
    self:root().location = location
end

---@private
---@return GuiLocation
function Toolbar:location()
    return self:root().location
end

---@public
function Toolbar:collapse()
    self:content():hide()
end

---@public
function Toolbar:expand()
    self:content():show()
end

---@public
function Toolbar:addSection()
    self:content():sectionList():addSection()
    self:resizeTablesInUnlock()
end

---@public
---@return number
function Toolbar:sectionsCount()
    local content = self:content()
    local sectionList = content:sectionList()
    local count = sectionList:count()
    return count
end

function Toolbar:lock()
    self:trimTables()
    Component.lock(self)
end

function Toolbar:unlock()
    self:resizeTablesInUnlock()
    Component.unlock(self)
end

---@public
function Toolbar:adjustSize()
    if not self:isLocked() then
        self:resizeTablesInUnlock()
    end
end

function Toolbar:trimTables()
    local minimumWidth = 0
    if self:sectionsCount() == 1 then
        minimumWidth = 1
    else
        minimumWidth = 2
    end
    self:resizeTables(minimumWidth, 1, 0)
end

---@private
function Toolbar:resizeTablesInUnlock()
    self:resizeTables(4, 1, 1)
end

function Toolbar:resizeTables(minimumWidth, minimumHeight, padding)
    local width = math.max(minimumWidth, self:lastOccupiedColumnIndex() + padding)
    self:adjustTables(width, minimumHeight, padding)
end

---@private
---@param width number
function Toolbar:adjustTables(width, minimumHeight, padding)
    for _, table in ipairs(self:tables()) do
        table:resize(width, minimumHeight, padding)
    end
end

---@return boolean
---@private
function Toolbar:isLocked()
    return self:child(ToolbarHeader):child(Unlock) ~= nil
end

---@public
---@return number
function Toolbar:lastOccupiedColumnIndex()
    local lastOccupiedColumnIndex = 0
    for _, table in ipairs(self:tables()) do
        lastOccupiedColumnIndex = math.max(lastOccupiedColumnIndex, table:lastOccupiedColumnIndex())
    end
    return lastOccupiedColumnIndex
end

---@private
---@return Table[]
function Toolbar:tables()
    local tables = {}
    for _, section in pairs(self:content():sectionList():sections()) do
        table.insert(tables, section:content():table())
    end
    return tables
end
